[TOC]

# 信号处理设计文档

## 一、信号概述
signal.c主要涉及的是进程的信号处理。信号是一种"软件中断"处理机制，它是操作系统向一个进程或者线程发送的一种异步通知，用于通知该进程或线程某种事件已经发生，需要做出相应的处理。

当进程接收到信号时，可以采取以下处理方式：

- 忽略信号：进程可以选择忽略某些信号，即不对该信号进行任何处理；
- 捕捉信号：进程可以注册一个信号处理函数来捕捉特定的信号，并在接收到该信号时执行相应的处理逻辑；
- 默认处理：如果进程没有注册信号处理函数且没有选择忽略信号，则系统会按照默认的处理方式来处理该信号。通常情况下，默认处理方式会导致进程终止或停止。

信号机制提供了一种处理异步时间的方法。例如用户在终端键盘上键入ctrl-C组合来终止一个程序的执行。该操作就会产生一个SIGINT信号，并被发送到当前的前台执行的进程中。

以下是signal模块运行的大致流程：

![image](./figs/signal.png)

其相关文件如下：

    ├─ signal.h //包括了信号和相关结构的的定义\
    ├─ signal.c
        ├─ issig(struct proc) //判断未决和阻塞状态
        ├─ psig(struct proc)  //信号处理
        ├─ signal(addr_sigret, int, sighandler_t)  //信号接收
        ├─ Kill(pid_t,int)  //向进程或进程组发送信号
        ├─ siggetmask(void)
        ├─ sigsetmask(int)
        ├─ sigblock(int)
        └─ sigunblock(int)
    ├─ syssignal.c  //包括了signal相关的系统调用
    └─signaltest.c  //编写了signal各项功能的测试


## 二、基本数据结构

这是相关的数据结构设计：

```c
typedef void (*sighandler_t)(int);
typedef void (*addr_sigret)(void);

typedef struct handlerinfo {
    //int signum
    int disposition;       //either DFL, IGN, or USERDEF
    sighandler_t handler;  //address of the user defined signal handler
}handlerinfo;
```

其中sighandler\_t定义为函数指针类型，对应的handler在signal结构中用来接收自定义的函数。执行调用的函数就是执行自定义的操作。

为保存信号，内核为每个进程维护了与信号相关的数据结构，主要包括以下变量：

```c
 //For signal handling
  uint sigpending;             // Pending signals
  uint sigblocked;             // blocked signals
  uint userdefed;              // User defined signals
  handlerinfo allinfo[NUMSIG]; // Information about signal handlers
  int justwoken;               // If process has just woken up after a sleep, etc.
  addr_sigret addr_sigreturn;  // Address of sigreturn as seen by the user
```

- sigpending：用于表示进程当前有哪些信号处于未决状态。当一个信号产生且未被处理时，其中对应的位会被设置为1;
- sigblocked：用于表示进程当前阻塞了哪些信号。每个信号都对应一个比特位，如果该位为1，则表示该信号被阻塞；为0则表示该信号未被阻塞；
- allinfo[NUMSIG]：用于表示当信号递达时应该执行的处理动作。每个信号都对应一个单元，可以是默认处理函数、忽略函数或用户自定义的处理函数；
- addr\_sigreturn：用于保存用户态下的sigreturn地址。

## 三、关键代码

当信号产生时，内核会执行以下操作来保存信号：

- 设置sigpending：在进程控制块的sigpending中，将对应信号位的值从0设置为1，表示该信号已经产生且处于未决状态；
- 检查sigblocked：内核会检查sigblocked，查看该信号是否被阻塞。如果被阻塞，则信号会保持在未决状态，不会被立即处理；
- 更新allinfo[NUMSIG]（如果需要）：如果进程为该信号设置了自定义处理函数，内核会更新allinfo中对应信号的处理函数指针为用户自定义的函数地址。

### 3.1 issig和psig

这一部分分别由 issig 和 psig 实现：

函数issig对sigpending和sigblocked进行判断，返回是否进行处理；

```c
int  
issig(struct proc *p){  
    return p->sigpending & (~(p->sigblocked));  
} 
```

当由内核态返回到用户态时，在返回的这个时候执行psig函数进行信号的检测和处理。函数psig应对各种信号做出相应的处理并在保存用户上下文后通过 jmptohandler函数跳转至信号处理程序，最后再转过去执行用户提供的handler方法，这里展示了psig函数的部分代码；

```c
int psig(struct proc *p)
{
    // 变量定义
    ......

    // 处理终止信号
    for (i = 0; i < 7; i++) {
        if (((p->sigpending & (1 << termsig[i])) == (1 << termsig[i])) &&
            ((p->sigblocked & (1 << termsig[i])) == 0) &&
            p->allinfo[termsig[i]].disposition == DFL) {
            if (termcalled == 0)
                term(p, &termcalled);
            p->sigpending &= ~(1 << termsig[i]);
        }
    }

    // 处理核心转储信号
    ......

    // 处理SIGSTOP信号
    if ((p->sigpending & (1 << 19)) == (1 << 19)) {
        stop(p, &stopcalled);
        p->sigpending &= ~(1 << 19);
    }

    
    ......

    // 处理仅有的SIGCONT信号
    if (((p->sigpending & (1 << 18)) == (1 << 18)) &&
        ((p->sigblocked & (1 << 18)) == 0) &&
        p->allinfo[SIGCONT].disposition == DFL) {
        cont(p);
        p->sigpending &= ~(1 << 18);
    }

    // 处理用户定义的信号处理程序
    if (p->userdefed == 0) {
            release(&proc->lock);
            return 1;
            }
        else {
            for (i = 1; i < 32; i++) {
                if ((p->userdefed & (1 << i)) == (1 << i))
                    break;
        }

        // 保存上下文到用户栈
        *(uint64 *)((p->trapframe->sp) - 8) = r_sstatus();
        ......

        p->sigpending &= ~(1 << i);
        release(&proc->lock);

        __sync_synchronize();

        // 跳转到用户定义的信号处理程序
        jmptohandler(p->allinfo[i].handler, r_sstatus(), 
                     r_scause(), p->trapframe->sp, r_sstatus());
    }
    return -1;
}
```

### 3.2 signal

signal.c中还包含了信号处理相关的其余函数：

signal函数负责接收信号并按照指定信号执行相应处理程序，其中returnfn 指定了系统调用 sigreturn 的用户地址，在系统调用sys\_signal中被调用。

```c
sighandler_t
signal(addr_sigret returnfn, int signum, sighandler_t handler) {

    // 错误状态返回以及变量初始化定义
    ......

    sighandler_t prev;
    if(p->allinfo[signum].disposition == USERDEF)
        prev = p->allinfo[signum].handler;
    else if(p->allinfo[signum].disposition == IGN)
         prev = SIG_IGN;
    else
        prev = SIG_DFL;
    
    acquire(&proc->lock);
    
    p->addr_sigreturn = NULL;
    if(handler == SIG_IGN) {
        //SIG_IGN处理
        if(p->allinfo[signum].disposition == USERDEF)
            p->userdefed = p->userdefed & (~(1 << signum));
        p->allinfo[signum].disposition = IGN;
    }
    else if(handler == SIG_DFL) {
        //SIG_DFL处理
        if(p->allinfo[signum].disposition == USERDEF)
            p->userdefed = p->userdefed & (~(1 << signum));
        p->allinfo[signum].disposition = DFL;
    }
    else {
        //自定义处理
        ......
    }

    release(&proc->lock);
    return prev;
}
```

### 3.3 Kill

Kill函数负责向进程或进程组发信号，其中sig为将要发送的信号，在系统调用sys\_Kill中被调用。

```c
int 
Kill(pid_t pid, int sig) {

    // 错误状态返回以及变量初始化定义
    ......
    
    for(p = proc; p < &proc[NPROC]; p++) {
        if(p->pid == 1)
                continue;
        if(minusone || p->pid == pid) {
            acquire(&proc->lock);

            if(sig == SIGKILL) {
                //SIGKILL处理
                ......
            }
            else if(sig == SIGCHLD && p->allinfo[SIGCHLD].disposition == DFL) {
                //SIGCHID处理
                ......
            }
            else if(sig == SIGCONT) {
                //SIGCONT处理
                ......
            }
            if(p->allinfo[sig].disposition != IGN && sig != 0) {
                p->sigpending = p->sigpending | (1 << sig);
            }

            release(&proc->lock);
            if(minusone == 0)
                break;
        }
    }
    return 0;
}
```

### 3.4 siggetmask和sigsetmask

siggetmask和sigsetmask是对掩码mask进行操作：

- siggetmask：取得现有阻塞信号掩码；
- sigsetmask：用给定信号掩码替换现有阻塞信号掩码；

分别在系统调用sys\_siggetmask和sys\_sigsetmask中被调用。

```c
int
siggetmask(void) {
    // 错误状态返回以及变量初始化定义
    ......

    acquire(&proc->lock);
    uint mask = p->sigblocked;
    release(&proc->lock);

    return mask;
}

int
sigsetmask(uint mask) {
    // 错误状态返回以及变量初始化定义
    ......

    acquire(&proc->lock);
    p->sigblocked = mask;
    release(&proc->lock);
    
    return mask;
}
```

### 3.5 sigblock和sigunblock

sigblock和sigunblock是对被阻塞信号掩码mask进行操作：

- sigblock：向被阻塞信号掩码中添加信号；
- sigunblock：解除给定信号的屏蔽；

分别在系统调用sys\_sigblock和sys\_sigunblock中被调用。

```c
int
sigblock(int sig) {
    if(sig <= 0 || sig > CURRNUM)
        return -1;
    if(sig == SIGKILL || sig == SIGSTOP || sig == SIGCONT)
        return 0;
    uint mask = siggetmask();
    mask = mask | (1 << sig);

    return (sigsetmask(mask));
}

int
sigunblock(int sig) {
    if(sig <= 0 || sig > CURRNUM)
        return -1;
    if(sig == SIGKILL || sig == SIGSTOP)
        return 0;
    uint mask = siggetmask();
    mask = mask & (~(1 << sig));

    return (sigsetmask(mask));
}
```
## 四、测试及运行结果

以下是运行测试程序signaltest.c的结果：

![image](./figs/test1.png)

在第一和第二部分中，首先使用Kill函数将进程的信号置为SIGKILL和SIGTERM，根据返回结果可得Kill发送成功。然后使用sigsetmask函数将SIGKILL、SIGSTOP、SIGCONT等信号屏蔽，并通过siggetmask取得当前的信号屏蔽码，根据返回结果可得屏蔽成功。

![image](./figs/test2.png)

在第三部分中，先使用sigblock函数将SIGTERM信号屏蔽，再使用sigunblock函数解除屏蔽，根据返回结果可得测试成功。
然后使用sigblock将SIGKILL和SIGSTOP信号依次屏蔽，根据返回结果可得屏蔽成功。最后使用sigblock进行任意屏蔽，测试其是否正确运行。
在第四部分中，程序具体全面的测试了Kill函数对各种信号的处理，包括SIGCHLD、SIGTSTP、SIGKILL等，以及任意赋值的尝试。

![image](./figs/test3.png)
![image](./figs/test4.png)

在第五部分中，对signal的系统调用进行了测试，包括但不限于对SIG_IGN、SIG_DFL等信号以及自定义中断的测试，同时也对错误赋值进行了多次测试。
至此测试程序结束，测试结果均符合预期。
