[TOC]

# XV6进程模块设计文档

## shutdown系统调用

操作系统需要提供安全地关闭系统的机制。shutdown系统调用实现了文件系统同步和硬件关机操作，确保系统可以安全、完整地终止运行。

功能实现：
```c
uint64
sys_shutdown(void)
{
  // 1. 文件系统同步
  begin_op();
  end_op();
  
  // 2. 硬件关机操作
  volatile uint32 *ctrl = (uint32*)SHUTDOWN_CTRL;
  volatile uint32 *addr = (uint32*)SHUTDOWN_ADDR;
  *ctrl = 1;
  __sync_synchronize();
  *addr = 0x5555;

  // 3. 关机失败处理
  printf("Shutdown failed\n");
  for(;;) {
    asm volatile("wfi");  // 等待中断指令
  }
  return 0;
}
```

设计亮点：
1. 采用三段式设计:
   - 文件系统同步确保数据完整性
   - 硬件关机尝试通过特定寄存器实现
   - 失败后进入低功耗等待状态

2. 安全保护机制:
   - 使用内存屏障保证指令顺序
   - 文件系统事务机制避免数据丢失
   - 失败处理防止系统不当运行

使用示例：
```bash
$ shutdown    # 系统开始关机流程
Syncing filesystem...
System halting...
```

这个设计保证了系统关机过程的安全性和可靠性，是系统管理中不可或缺的组件。


## PS系统调用
系统调用设计：PS进程状态查看命令

在操作系统中,用户需要一种方式来查看和监控系统中运行的进程。PS命令作为最基本的进程管理工具,能够展示所有活跃进程的运行状态,帮助用户了解系统运行情况并进行故障诊断。此系统调用将为用户提供进程查看功能。

系统调用通过遍历进程表来获取所有非UNUSED状态的进程信息。对每个进程,显示其PID、名称、状态和优先级等关键信息。采用表格形式展示,确保输出格式规范整齐。具体代码实现如下:

```c
uint64 sys_ps(void) {
  struct proc *p;
  printf("PID\tName\tState\tPriority\n");
  for(p = proc; p < &proc[NPROC]; p++){
    if(p->state != UNUSED){
      printf("%d\t%s\t%d\t%d\n", p->pid, p->name, p->state, p->priority);
    }
  }
  return 0;
}
```

我们采用了直接遍历方式,保证了代码的简洁性和执行效率。通过系统调用机制确保了进程信息访问的安全性,用户无法直接访问内核数据结构。代码设计预留了扩展空间,比如进程优先级显示为后续调度优化做准备。表格化的输出格式也方便后续添加更多进程属性的显示。

使用示例：
用户可以在终端中直接输入ps命令查看当前系统中的进程状态：
```bash
$ ps
PID     Name    State   Priority
1       init    1       5
2       sh      2       5
3       ps      3       5
```

这样的设计既满足了基本的进程查看需求,又为后续功能扩展预留了空间,是一个简洁而实用的系统调用实现。


## TOP系统调用

系统调用设计：TOP系统资源监控命令

在进程管理中,除了基本的进程状态查看,还需要一个更全面的系统资源监控工具。TOP命令提供了系统运行时间、CPU使用率、内存占用等关键指标的实时监控,是系统管理和性能分析的重要工具。

TOP命令首先清屏以提供清晰视图,然后显示系统运行时间。对每个进程,计算并展示其CPU使用率、内存占用、运行时间、优先级、状态等信息。具体实现如下:

```c
uint64 
sys_top(void)
{
  struct proc *p;
  uint64 total_mem = PHYSTOP - KERNBASE;  // 总内存
  uint64 used_mem = 0;                    // 已用内存
  uint64 current_ticks;                   // 当前时钟
  
  acquire(&tickslock);
  current_ticks = ticks;
  release(&tickslock);
  
  printf("\x1b[2J\x1b[H");               // 清屏
  printf("=== XV6 TOP ===\n");
  printf("Uptime: %ld seconds\n", current_ticks/10);
  printf("PID\tCPU%%\tMEM\tTIME\tPRI\tSTATE\t\tNAME\n");
  
  // 计算内存使用
  for(p = proc; p < &proc[NPROC]; p++) {
    if(p->state != UNUSED) {
      used_mem += p->sz;
    }
  }
  
  // 显示进程详情
  for(p = proc; p < &proc[NPROC]; p++) {
    if(p->state != UNUSED) {
      int cpu_usage = (p->cputicks * 100) / (ticks - p->ctime + 1);
      int mem_percent = (p->sz * 100) / total_mem;
      int runtime = (ticks - p->ctime) / 10;
      
      printf("%d\t%d%%\t%d%%\t%d\t%d\t%s\t%s\n",
             p->pid, cpu_usage, mem_percent, runtime,
             p->priority, state2str(p->state), p->name);
    }
  }
  return 0;
}
```

我们通过清屏机制确保显示清晰,采用tick计数评估CPU使用率,并通过进程大小计算内存占用。状态显示采用对齐格式,保证了良好的可读性。此外,预留了CPU使用率优化空间,为后续性能分析提供基础。

使用效果：
```bash
=== XV6 TOP ===
Uptime: 100 seconds
PID     CPU%    MEM     TIME    PRI     STATE           NAME
1       5%      2%      100     3       run             init
2       10%     5%      98      3       sleep           sh
3       20%     8%      50      4       runble          top
```

这样的实现既保持了xv6的简洁风格,又提供了Linux top命令的核心功能。

## vfork
首先，子进程和父进程不共用用户的页表，但有相同的虚拟地址空间（这里指结构相同）和有相同的物理地址，对此修改了vm.c文件，加入了新的函数uvmcopy_v(),是修改uvmcopy()得来。
子进程会睡眠父进程。
```c
int
vfork(void)
{
  int i, pid;
  struct proc *np;
  struct proc *p = myproc();

  // Allocate process.
  if((np = allocproc()) == 0){
    return -1;
  }

  // Copy user memory from parent to child.
  if(uvmcopy_v(p->pagetable, np->pagetable, p->sz) < 0){
    freeproc(np);
    release(&np->lock);
    return -1;
  }
  np->sz = p->sz;
  
  // copy saved user registers.
  *(np->trapframe) = *(p->trapframe);

  // Cause fork to return 0 in the child.
  np->trapframe->a0 = 0;

  // increment reference counts on open file descriptors.
  for(i = 0; i < NOFILE; i++)
    if(p->ofile[i])
      np->ofile[i] = filedup(p->ofile[i]);
  np->cwd = idup(p->cwd);
  safestrcpy(np->name, p->name, sizeof(p->name));

  pid = np->pid;

  release(&np->lock);

  acquire(&wait_lock);
  np->parent = p;
  
  acquire(&np->lock);
  np->state = RUNNABLE;
  np->v_flag = 1;
  release(&np->lock);
  acquire(&p->lock);  //DOC: sleeplock1
  release(&wait_lock);

  // Go to sleep.
  p->chan = p;
  p->state = SLEEPING;

  sched();
  // Tidy up.
  p->chan = 0;

  // Reacquire original lock.
  release(&p->lock);
  return pid;
}
```


uvmcopy_v(),uvmcopy()的不同是uvmcopy_v()并不会再分配不关于页表的物理空间，而是让页表指向同一物理空间，而uvmcopy()则是分配物理空间且复制父进程页表的相同虚拟地址对应物理空间的值。通过uvmcopy_v()该方法实现物理空间的共享。

```c
int
uvmcopy(pagetable_t old, pagetable_t new, uint64 sz)
{
  pte_t *pte;
  uint64 pa, i;
  uint flags;
  char *mem;

  for(i = 0; i < sz; i += PGSIZE){
    if((pte = walk(old, i, 0)) == 0)
      panic("uvmcopy: pte should exist");
    if((*pte & PTE_V) == 0)
      panic("uvmcopy: page not present");
    pa = PTE2PA(*pte);
    flags = PTE_FLAGS(*pte);
    if((mem = kalloc()) == 0)
      goto err;
    memmove(mem, (char*)pa, PGSIZE);
    if(mappages(new, i, PGSIZE, (uint64)mem, flags) != 0){
      kfree(mem);
      goto err;
    }
  }
  return 0;

 err:
  uvmunmap(new, 0, i / PGSIZE, 1);
  return -1;
}
int
uvmcopy_v(pagetable_t old, pagetable_t new, uint64 sz)
{
  pte_t *pte;
  uint64 pa, i;
  uint flags;

  for(i = 0; i < sz; i += PGSIZE){
    if((pte = walk(old, i, 0)) == 0)
      panic("uvmcopy: pte should exist");
    if((*pte & PTE_V) == 0)
      panic("uvmcopy: page not present");
    pa = PTE2PA(*pte);
    flags = PTE_FLAGS(*pte);
    mappages(new, i, PGSIZE, (uint64)pa, flags);

  }
  return 0;
}
```

使用变量验证vfork的过程中，发现不给变量加volatile会被编译器的寄存器优化给优化，即寄存器保存变量没有写回内存导致。

```c
#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"
int main (int argc,char *argv[])
{ int p;
volatile int a= 0;
int b = 0;

   if((p=vfork())==0){
    a+=100;
    b+=2;
    printf("zhi a= %d b = %d\n",a,b);
    exit(0);
   }
   else{
    a += 2;
    b+=1;
    printf("fa a = %d,b = %d\n",a,b);
   }

//printf("This is my own app and cpu id is:%d!\n",getcpuid());
exit(0);
}
```
下面是结果

![Local Image](./figs/vfork.png)




对应可能的问题，并没有实现写时复制等。处理了回收内存的问题，因为是共用的内存物理地址。